from itertools import accumulate, combinations, product
import re
import inspect
from artiruno.preorder import PreorderedSet, Relation, IC, LT, EQ, GT
from artiruno.util import cmp, choose2

class Jump(Exception):
    def __init__(self, value):
        self.value = value

class Abort(Exception): pass

avda_doc = 'As :func:`vda`, but accepts an asynchronous ``asker``.'

async def avda(
        criteria, alts = None, asker = None, find_best = None,
        max_dev = 2, allowed_pairs_callback = lambda x: None):
    '''Conduct verbal decision analysis.

    :param criteria: An iterable of iterables specifying the levels of each criterion. Levels can be any hashable object, but are typically strings. Within a criterion, we assume that later levels are better.
    :param alts: An iterable of the alternatives; that is, the specific items that can be decided among. Each alternative is represented as an iterable of criterion levels, listed in the same order as the criteria. If ``alts`` is :data:`py:None`, we use the entire item space; that is, the set of all possible items.
    :param asker: A callable object ``f(a, b)`` that returns a :class:`Relation` (other than :const:`IC <Relation.IC>`) for ``a`` and ``b``; greater elements represent greater preference. Use :func:`vda` if the asker is synchronous and :func:`avda` if it's asynchronous.
    :param find_best: An integer. If set, Artiruno will aim to identify the top ``find_best`` items and stop there. Otherwise, Artiruno will try to compare all the alternatives.
    :param max_dev: The maximum number of criteria on which hypothetical items can deviate from the reference item when asking the user to make choices. It's summed across both items; e.g., ``max_dev = 5`` allows 4 deviant criteria compared to 1 deviant criterion, or 3 compared to 2.
    :param allowed_pairs_callback: Called on ``allowed_pairs`` for each iteration of the outermost loop.

    :returns: A :class:`PreorderedSet`. Its elements will be a subset of the item space and a superset of ``alts``.'''

    criteria, alts, prefs = _setup(criteria, alts)
    assert 2 <= max_dev <= 2*len(criteria)
    if find_best:
        assert 1 <= find_best <= len(alts)

    async def get_pref(a, b):
        add_items(criteria, prefs, [a, b])
        if (rel := prefs.cmp(a, b)) == IC:
            rel = await asker(a, b)
            prefs.learn(a, b, rel)
        return rel

    def dev_from_ref(dev_criteria, vector):
        # Return a vector in the item space that deviates from the
        # best possible item on the given criteria with the given
        # values.
        return tuple(
           vector[i] if i in dev_criteria else c[-1]
           for i, c in enumerate(criteria))

    def num_item(item):
        # Represent criterion values as integers. This prevents us
        # from behaving differently depending on the default sort
        # order of the criterion values. (We're still sensitive to the
        # order of criteria, though.)
        return tuple(criteria[i].index(v) for i, v in enumerate(item))

    for allowed_pairs in (l[::-1] for l in accumulate(
           [(big, small)] + ([] if big == small else [(small, big)])
           for deviation in range(1, max_dev)
           for big in range(deviation, deviation // 2, -1)
           for small in [deviation + 1 - big]
           if small <= len(criteria))):

        allowed_pairs_callback(allowed_pairs)

        to_try = set(choose2(sorted(alts, key = num_item)))

        while True:

            if find_best and len(prefs.maxes(alts)) >= find_best:
                return prefs

            # Don't ask about pairs we already know.
            to_try = {x for x in to_try if prefs.cmp(*x) == IC}

            if find_best:
                # Don't compare alternatives that can't be in the
                # requested `extreme` set.
                not_best = {x
                    for x in alts
                    if sum(prefs.cmp(x, a) == LT for a in alts) >=
                       find_best}
                to_try = {(a, b)
                    for a, b in to_try
                    if a not in not_best and b not in not_best}

            if not to_try:
                if any(prefs.cmp(a, b) == IC for a, b in choose2(alts)):
                    break
                return prefs

            a, b = max(to_try, key = lambda pair:
                (num_item(pair[0]), num_item(pair[1])))
            to_try.remove((a, b))

            cs = frozenset({ci
                for ci in range(len(criteria))
                if a[ci] != b[ci]})
            try:
                async def f(rel, cs1, cs2):
                    if not cs1:
                        raise Jump(rel)
                    for size1, size2 in allowed_pairs[:
                            # In the topmost call of `f`, force use of the
                            # new allowed_pairs, to save pointless iterations.
                            2 if len(cs1) == len(criteria) else None]:
                        if (len(cs1) - size1 < 0 or len(cs2) - size2 < 0 or
                                (len(cs1) - size1 == 0) != (len(cs2) - size2 == 0)):
                            continue
                        for c1 in combinations(sorted(cs1), size1):
                            for c2 in combinations(sorted(cs2), size2):
                                p = await get_pref(dev_from_ref(c1, a), dev_from_ref(c2, b))
                                if rel == EQ or p in (EQ, rel):
                                    await f(rel or p, cs1.difference(c1), cs2.difference(c2))
                await f(EQ, cs, cs)
            except Jump as j:
                prefs.learn(a, b, j.value)
            except Abort:
                return prefs

    return prefs

# Define `vda` as a a synchronous version of `avda`.
exec(
  re.sub(r'\basync ', '',
  re.sub(r'\bawait ', '',
  re.sub('def avda', 'def vda',
  ''.join(inspect.getsourcelines(avda)[0])))))
avda.__doc__ = avda_doc

def _setup(criteria, alts = None, find_best = None):
    # Some initial VDA logic put into its own function so it can be
    # tested separately.

    criteria = tuple(map(tuple, criteria))
    assert len(criteria)
    assert all(
        len(c) > 0 and len(c) == len(set(c))
        for c in criteria)

    if alts is None:
        alts = tuple(product(*criteria))
    else:
        alts = tuple(map(tuple, alts))
        assert all(
            len(a) == len(criteria) and all(
                a[i] in criteria[i]
                for i in range(len(a)))
            for a in alts)
        assert len(alts) == len(set(alts))

    # Define the user's preferences as a preorder, with `a < b` if `b`
    # is preferred to `a`, and `a` and `b` incomparable if the user's
    # preference isn't yet known.
    prefs = PreorderedSet()
    add_items(criteria, prefs, alts)

    return criteria, alts, prefs

def add_items(criteria, prefs, items):
    # Enforce the assumption that on any single criterion, bigger
    # values are better.
    for x in set(items) - prefs.elements:
        prefs.add(x)
        for a in prefs.elements - {x}:
            cmps = [
                Relation.cmp(
                    criteria[ci].index(x[ci]),
                    criteria[ci].index(a[ci]))
                for ci in range(len(criteria))]
            if not (LT in cmps and GT in cmps):
                # One item dominates the other. (We know that `cmps`
                # isn't all EQ because `x` and `a` are different.)
                prefs.learn(x, a, LT if LT in cmps else GT)



import itertools

def cmp(a, b):
    "As Python 2's :func:`py2:cmp`."
    return (a > b) - (a < b)

def choose2(x):
    'Shortcut for ``itertools.combinations(x, 2)``.'
    return itertools.combinations(x, 2)



import pytest

def pytest_addoption(parser):
    parser.addoption('--slow',
        action = 'store_true', default = False,
        help = 'run slow tests')

def pytest_configure(config):
    config.addinivalue_line('markers', 'slow: mark test as slow to run')

def pytest_collection_modifyitems(config, items):
    # Skip slow tests unless `--slow` is provided.
    if config.getoption('--slow'):
       return
    for item in items:
        if 'slow' in item.keywords:
            item.add_marker(pytest.mark.skip(reason =
                '--slow not provided'))

@pytest.fixture
def run_slow_tests(request):
    return request.config.getoption('--slow')



import itertools, enum
from collections import Counter
from artiruno.util import cmp, choose2

class Relation(enum.Enum):
    'An :class:`enum.Enum` representing the order relation between two objects.'

    IC = None  #: Incomparable (true relation not yet known)
    LT = -1    #: Less than
    EQ = 0     #: Equal or equivalent to
    GT = 1     #: Greater than

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name

    def __bool__(self):
        ':const:`LT` and :const:`GT` are true. :const:`IC` and :const:`EQ` are false.'
        return bool(self.value)

    def __neg__(self):
        'Swap :const:`LT` and :const:`GT`. Return :const:`IC` or :const:`EQ` unchanged.'
        return (
            LT if self == GT else
            GT if self == LT else
            self)

    @classmethod
    def cmp(cls, a, b):
        "Return the :class:`Relation` between ``a`` and ``b`` corresponding to Python's built-in comparison operators."
        return cls(cmp(a, b))

IC, LT, EQ, GT = Relation.IC, Relation.LT, Relation.EQ, Relation.GT

class ContradictionError(Exception):
    'Represents an attempt to build an inconsistent order.'
    def __init__(self, k, was, claimed):
        super().__init__('{}: known to be {}, now claimed to be {}'.format(k, was, claimed))

class PreorderedSet:
    '''A set equipped with a `preorder <https://en.wikipedia.org/wiki/Preorder>`_. Elements can be added to the set, and the order can be updated to make previously incomparable elements comparable.

    :param elements: An iterable of hashable objects to be ordered.
    :param relations: An iterable of triples ``(a, b, rel)``, where ``a`` and ``b`` are objects in ``elements``, and ``rel`` is a :class:`Relation`. By default, all elements are incomparable to each other.
    :param raw_relations: Used internally.'''

    def __init__(self, elements = (), relations = (), raw_relations = None):
        self.elements = set(elements)
        self.relations = raw_relations or {(a, b): IC
            for a, b in choose2(sorted(self.elements))}
        for a, b, rel in relations:
            self.learn(a, b, rel)

    def __repr__(self):
        return f'PreorderedSet({self.elements!r}, {self.relations!r})'

    def copy(self):
        'Return a copy of the object.'
        return type(self)(self.elements.copy(),
            raw_relations = self.relations.copy())

    def add(self, x):
        "Add ``x`` to the set; a no-op if it's already there. ``x`` is initially incomparable to all other elements."
        if x in self.elements:
            return
        for a in self.elements:
            self.relations[(a, x) if a < x else (x, a)] = IC
        self.elements.add(x)

    def cmp(self, a, b):
        'Return the :class:`Relation` between ``a`` and ``b``.'
        if a == b:
            assert a in self.elements and b in self.elements
            return EQ
        return self.relations[a, b] if a < b else -self.relations[b, a]

    def extreme(self, n, among = None, bottom = False):
        '''Return the top-``n`` subset (or bottom-``n`` subset, if ``bottom`` is true) of the items in the set ``among``, or of the whole set if ``among`` is not provided.

        We define the *top-n subset* of a preordered set ``S`` to be the set of all elements ``x ∈ S`` such that

        - ``x`` is comparable to all elements of ``S``, and
        - there are at most ``n - 1`` distinct elements ``a ∈ S`` such that ``a > x``.

        We define the bottom-``n`` subset similarly, with the inequality in the other direction. Notice that the top-``n`` subset may contain more or less than ``n`` items.'''

        return frozenset(x
            for x in among or self.elements
            for cmps in [Counter(
                self.cmp(x, a) for a in among or self.elements)]
            if cmps[IC] == 0 and cmps[GT if bottom else LT] < n)

    def maxes(self, among = None):
        'Return all the maxima among the items in ``among``, or the whole set if ``among`` is not provided. The maxima are defined as the top-1 subset, per :meth:`extreme`.'
        return self.extreme(1, among, bottom = False)
    def mins(self, among = None):
        'As ``maxes``, but for minima.'
        return self.extreme(1, among, bottom = True)

    def _set(self, a, b, rel):
        # Return true if a change was made.
        assert rel in (LT, EQ, GT)

        if a == b:
             assert a in self.elements
             if rel != EQ:
                 raise ContradictionError((a, b), EQ, rel)
             return False

        k = (a, b)
        if b < a:
           k, rel = (b, a), -rel

        if self.relations[k] == IC:
            self.relations[k] = rel
            return True
        elif self.relations[k] == rel:
            return False
        else:
            raise ContradictionError(k, self.relations[k], rel)

    def learn(self, a, b, rel):
        '''Update the ordering such that ``a`` and ``b`` are related by ``rel``, a :class:`Relation` other than :const:`IC <Relation.IC>`. Any other relations that can be inferred by transitivity will be added to the set.

        Return a list of pairs that were actually updated. If the relation between ``a`` and ``b`` was already known, the list will be empty. If we made inferences for other pairs, these pairs will be included in the list.

        Raise :class:`ContradictionError` if the new relation isn't consistent with the preexisting relations (other than incomparability).'''

        if not self._set(a, b, rel):
            return []
        # Use a modification of Warshall's algorithm to update the transitive
        # closure.
        # https://web.archive.org/web/2013/https://cs.winona.edu/lin/cs440/ch08-2.pdf
        changed = [(a, b) if a < b else (b, a)]
        for k, i, j in itertools.product((a, b), self.elements, self.elements):
            r1, r2 = self.cmp(i, k), self.cmp(k, j)
            if (r1 != IC and r2 == EQ) or (r1 == LT == r2):
                if self._set(i, j, r1):
                    changed.append((i, j) if i < j else (j, i))
        return changed

    def get_subset(self, elements):
        'Return a new :class:`PreorderedSet` that contains only the requested ``elements``.'

        elements = frozenset(elements)
        assert elements.issubset(self.elements)
        return type(self)(elements, raw_relations = {(a, b): r
            for (a, b), r in self.relations.items()
            if {a, b}.issubset(elements)})

    def summary(self, namer = str):
        'Describe all the relations with a string like "A<B C<D E=F". ``namer`` should be a callback that returns a name for an element, as a string.'
        return ' '.join(
            '{}{}{}'.format(
                namer(a), {EQ: '=', LT: '<'}[rel], namer(b))
            for a, b, rel in sorted(
                (b, a, LT) if rel == GT else (a, b, rel)
                for (a, b), rel in self.relations.items()
                if rel != IC))

    def graph(self, namer = str):
        'Return the set represented as a :class:`graphviz.Source` object. Requires the Python package ``graphviz``. ``namer`` should be a callback that returns a name for an element, as a string.'

        import subprocess
        import graphviz

        g = graphviz.Digraph()

        # Collect groups of equivalent items into nodes.
        nodes = []
        for a in sorted(self.elements):
            for node in nodes:
                if self.cmp(a, node[0]) == EQ:
                    node.append(a)
                    break
            else:
                nodes.append([a])

        # Add the nodes to the graph.
        def node_repr(node):
            return ' / '.join(map(namer, sorted(node)))
        for node in nodes:
            g.node(node_repr(node))

        # Add edges.
        for a, b in itertools.product(nodes, nodes):
            if self.cmp(a[0], b[0]) == LT:
                g.edge(node_repr(b), node_repr(a))

        # Render to Graphviz source code, transform to the transitive
        # reduction with `tred`, and redigest the graph.
        return graphviz.Source(subprocess
            .Popen(['tred'], stdin = subprocess.PIPE, stdout = subprocess.PIPE)
            .communicate(g.source.encode('UTF-8'))[0].decode('UTF-8'))



from artiruno import (
    PreorderedSet, Relation, IC, LT, EQ, GT, ContradictionError, cmp, choose2)
import pytest

def test_simple():
    x = PreorderedSet(range(2))
    assert x.summary() == ""
    x.learn(0, 1, LT)
    assert x.summary() == "0<1"
    with pytest.raises(ContradictionError):
        x.learn(1, 0, LT)
    with pytest.raises(ContradictionError):
        x.learn(0, 0, LT)

    x = PreorderedSet(range(2))
    x.learn(1, 0, EQ)
    assert x.summary() == "0=1"

def test_trans():
    x = PreorderedSet("abc")
    x.learn("a", "b", LT)
    x.learn("c", "b", GT)

    y = PreorderedSet("abc", [["a", "b", LT], ["c", "b", GT]])

    for s in x, y:
        assert s.summary() == "a<b a<c b<c"
        with pytest.raises(ContradictionError):
            s.learn("c", "a", LT)

def test_long_trans():
    x = PreorderedSet(set(range(20)) - {10})
    for s in list(range(10)), list(range(11, 20)):
        for a, b in zip(s, s[1:]):
            x.learn(a, b, LT)
    assert x.cmp(8, 9) == LT
    assert x.cmp(11, 12) == LT
    assert x.cmp(9, 11) == IC
    assert x.cmp(0, 19) == IC
    y = x.copy()

    x.learn(9, 11, LT)
    assert x.cmp(9, 11) == LT
    assert x.cmp(0, 19) == LT

    y.learn(9, 11, EQ)
    assert y.cmp(9, 11) == EQ
    assert y.cmp(0, 19) == LT

def test_complex():
    x = PreorderedSet("abcwxyz")

    x.learn("a", "b", LT)
    x.learn("c", "b", GT)
    x.learn("b", "w", EQ)
    assert x.summary() == "a<b a<c a<w b<c b=w w<c"
    was = x.summary()

    x.learn("x", "y", LT)
    x.learn("y", "z", LT)
    assert x.summary() == was + " x<y x<z y<z"

    y = x.copy()

    x.learn("b", "y", LT)
    assert x.summary() == "a<b a<c a<w a<y a<z b<c b=w b<y b<z w<c w<y w<z x<y x<z y<z"

    y.learn("b", "y", EQ)
    assert y.summary() == "a<b a<c a<w a<y a<z b<c b=w b=y b<z w<c w=y w<z x<b x<c x<w x<y x<z y<c y<z"

    return x  # For `graph_example`.

def test_big():
    x = PreorderedSet(range(25))
    for a, b, r in [[6, 16, EQ], [16, 5, LT], [21, 11, EQ], [8, 10, EQ], [11, 20, LT], [8, 3, EQ], [1, 18, EQ], [2, 3, GT], [0, 2, LT], [12, 23, EQ], [1, 20, EQ], [23, 1, GT], [16, 13, LT], [14, 6, EQ], [1, 22, GT], [13, 2, GT], [24, 5, LT], [16, 3, EQ], [7, 16, EQ], [19, 16, EQ], [4, 17, LT], [4, 0, GT], [3, 7, EQ], [15, 16, EQ]]:
        x.learn(a, b, r)
    assert x.summary() == '0<2 0<4 0<13 0<17 1<12 1=18 1=20 1<23 2<13 3<2 3<5 3=6 3=7 3=8 3=10 3<13 3=14 3=15 3=16 3=19 4<17 6<2 6<5 6=7 6=8 6=10 6<13 6=14 6=15 6=16 6=19 7<2 7<5 7=8 7=10 7<13 7=14 7=15 7=16 7=19 8<2 8<5 8=10 8<13 8=14 8=15 8=16 8=19 10<2 10<5 10<13 10=14 10=15 10=16 10=19 11<1 11<12 11<18 11<20 11=21 11<23 12=23 14<2 14<5 14<13 14=15 14=16 14=19 15<2 15<5 15<13 15=16 15=19 16<2 16<5 16<13 16=19 18<12 18=20 18<23 19<2 19<5 19<13 20<12 20<23 21<1 21<12 21<18 21<20 21<23 22<1 22<12 22<18 22<20 22<23 24<5'

def test_extrema():
    x = PreorderedSet(range(4))
    x.learn(0, 1, LT)
    x.learn(1, 2, LT)
    assert x.mins() == x.maxes() == set()
    x.learn(2, 3, EQ)
    assert x.mins() == {0}
    assert x.maxes() == {2, 3}
    assert x.maxes(among = {0, 1, 3}) == {3}
    assert x.extreme(n = 2) == {2, 3}
    assert x.extreme(n = 3) == {1, 2, 3}

def test_extreme_n():
    x = PreorderedSet(l + str(i)
        for l in "wxyz"
        for i in range(3))
    for a, b in choose2(x.elements):
        x.learn(a, b, Relation.cmp(a[0], b[0]))

    assert (x.maxes() == x.extreme(2) == x.extreme(3) ==
        {"z0", "z1", "z2"})
    assert (x.extreme(4) == x.extreme(5) == x.extreme(6) ==
        {"z0", "z1", "z2", "y0", "y1", "y2"})
    def xb(n):
        return x.extreme(n, bottom = True)
    assert (x.mins() == xb(2) == xb(3) ==
        {"w0", "w1", "w2"})
    assert (xb(4) == xb(5) == xb(6) ==
        {"w0", "w1", "w2", "x0", "x1", "x2"})

def test_get_subset():
    x = PreorderedSet(("a", "b1", "b2", "c", "d"), (
        ("a", "b1", LT), ("a", "b2", LT),
        ("b1", "c", LT), ("b2", "c", LT),
        ("c", "d", LT)))
    y = x.get_subset(("a", "b2", "d"))
    assert x.summary() == "a<b1 a<b2 a<c a<d b1<c b1<d b2<c b2<d c<d"
    assert y.elements == {"a", "b2", "d"}
    assert y.summary() == "a<b2 a<d b2<d"

def test_add():
    x = PreorderedSet()
    x.add('a')
    x.add('z')
    assert x.cmp('a', 'z') == IC
    x.learn('a', 'z', LT)
    assert x.cmp('a', 'z') == LT
    x.add('f')
    x.learn('f', 'a', LT)
    assert x.cmp('f', 'a') == LT
    assert x.cmp('f', 'z') == LT

def test_learn_return_value():
    x = PreorderedSet((1, 2, 3))
    assert x.learn(1, 2, LT) == [(1, 2)]
    assert x.learn(2, 3, LT) == [(2, 3), (1, 3)]
    assert x.learn(1, 2, LT) == []


def graph_example():
    x = test_complex()
    x.learn("a", "x", LT)
    x.graph().render("/tmp/graph", format = 'png', view = True)

if __name__ == '__main__':
    graph_example()



# Set up the test suite for running it with Pyodide.
# Command-line arguments are passed to `pytest`.
# N.B. `--slow` on Firefox takes about 15 minutes on my fairly beefy
# laptop, vs. less than a tenth of that with native CPython.

import sys, tempfile, subprocess, re
from pathlib import Path

pyodide_version = re.search('pyodide/(v[^/]+)',
    Path('webi.html').read_text()).group(1)

td = Path('/tmp/artiruno_pyodide_testing_SNtl1aBcvhoD5PO8upr4')
td.mkdir(exist_ok = True)
subprocess.check_call(['tar',
    '--create', '--auto-compress',
    '--file', td / 'artiruno_for_pyodide.tar.gz',
    '--exclude', '__pycache__',
    *'artiruno conftest.py examples pytest.ini tests setup.py'.split()])

page = '''
    <!DOCTYPE html>
    <script src="https://cdn.jsdelivr.net/pyodide/PYV/full/pyodide.js"></script>
    <title>Artiruno Pyodide testing</title>
    <script>
    async function main()
       {let pyodide = await loadPyodide(
           {indexURL: "https://cdn.jsdelivr.net/pyodide/PYV/full/"})
        await pyodide.loadPackage('pytest')
        await pyodide.loadPackage('micropip')
        await pyodide.runPythonAsync(`
            import pyodide
            pyodide.webloop.WebLoop.close = lambda *a, **kw: True
              # Work around https://github.com/pyodide/pyodide/issues/2221
            import micropip
            await micropip.install('pytest-asyncio')
            from pyodide.http import pyfetch
            await (await pyfetch('artiruno_for_pyodide.tar.gz')).unpack_archive()`)
        console.log('pytest exit code: ', pyodide.runPython(`
            import pytest
            pytest.main(['-s', '--verbose', '--color=no', *ARGS])`))}
    main()
    </script>'''
(td / 'page.html').write_text(page
    .replace('PYV', pyodide_version)
    .replace('ARGS', repr(sys.argv[1:])))

print(f'''Now launch a test server with
    python3 -m http.server --bind 127.0.0.1 --dir {td}
open the page with
    "$BROWSER" http://0.0.0.0:8000/page.html
and check the browser console for pytest's output.''')



import os, sys
sys.path.insert(0, os.path.abspath('..'))

import artiruno
release = artiruno.__version__

extensions = [
     'sphinx.ext.autodoc',
     'sphinx.ext.intersphinx']

exclude_patterns = ['_build']

smartquotes = False
nitpicky = True

intersphinx_mapping = dict(
    py = ('https://docs.python.org/3/', None),
    py2 = ('https://docs.python.org/2/', None),
    graphviz = ('https://graphviz.readthedocs.io/en/stable', None))



import random, itertools, inspect, asyncio
from collections import Counter
import artiruno
from artiruno import Relation, IC, LT, EQ, GT, vda, avda, cmp, choose2
import pytest

def test_assumptions():
    '''Test the preferences that we should assume purely from the
    criteria definitions, without having asked the user anything
    yet.'''

    *_, prefs = artiruno.m.vda._setup(
        criteria = ['abcd', ('bad', 'okay', 'good')])
    assert prefs.cmp(('a', 'bad'), ('c', 'bad')) == LT
    assert prefs.cmp(('a', 'okay'), ('a', 'good')) == LT
    assert prefs.cmp(('d', 'good'), ('a', 'bad')) == GT
    assert prefs.cmp(('c', 'okay'), ('b', 'bad')) == GT
    assert prefs.cmp(('b', 'good'), ('c', 'okay')) == IC

    *_, prefs = artiruno.m.vda._setup(
        criteria = [range(3) for _ in range(3)],
        alts = ((1, 1, 1), (2, 1, 2)))
    assert prefs.cmp((1, 1, 1), (2, 1, 2)) == LT

def test_appendixD():
    '''Appendix D of Larichev and Moshkovich (1995).'''

    criteria = [(3,2,1)] * 3
      # 3 is worst and 1 is best.
    alts = [(1,2,3), (2,3,1), (3,1,2)]
    dm_ranking = [(2,1,1), (1,1,2), (1,2,1), (1,3,1), (3,1,1), (1,1,3)]
      # The first element is the best, so we negate `cmp` in the
      # asker.
    def asker(a, b):
        return -Relation.cmp(dm_ranking.index(a), dm_ranking.index(b))

    Proposal1, Proposal2, Proposal3 = alts

    for goal in ('find_best', 'rank_alts', 'rank_space'):
        prefs = vda(
            criteria = criteria,
            alts = (None if goal == 'rank_alts' else alts),
            asker = asker,
            find_best = goal == 'find_best' and 1)
        assert prefs.maxes(among = alts) == {Proposal2}
        if goal == 'rank_alts':
            for v1, v2 in choose2(dm_ranking):
                assert prefs.cmp(v1, v2) == -Relation.cmp(
                    dm_ranking.index(v1), dm_ranking.index(v2))
        if goal != 'find_best':
            assert prefs.cmp(Proposal2, Proposal1) == GT
            assert prefs.cmp(Proposal2, Proposal3) == GT
            assert prefs.cmp(Proposal3, Proposal1) == GT

def test_simple_strings():

    criteria = [['bad', 'good'], ['expensive', 'cheap']]
    alts = [
        ('bad', 'cheap'), ('good', 'expensive'),
        ('bad', 'expensive')]

    def asker(a, b):
       # We prefer good to bad, but when that criterion is the same,
       # we prefer cheap to expensive.
        if a[0] == b[0]:
            if a[1] == b[1]:
                return EQ
            elif a[1] == 'cheap':
                return GT
            else:
                return LT
        elif a[0] == 'good':
            return GT
        else:
            return LT

    for find_best in (1, 2, None):
        prefs = vda(
            criteria = criteria,
            alts = alts if find_best else None,
            asker = asker,
            find_best = find_best)
        assert prefs.maxes(among = alts) == {('good', 'expensive')}
        if find_best == 2:
            assert prefs.extreme(2, among = alts) == {
                ('good', 'expensive'), ('bad', 'cheap')}
        elif not find_best:
            assert prefs.maxes() == {('good', 'cheap')}
            ranking = (
                ('bad', 'expensive'), ('bad', 'cheap'),
                ('good', 'expensive'), ('good', 'cheap'))
            for (ai, a), (bi, b) in choose2(enumerate(ranking)):
                assert prefs.cmp(a, b) == Relation.cmp(ai, bi)

async def test_async():
    criteria = ((0, 1), (0, 1))
    alts = ((0, 1), (1, 0))
    l = []

    async def f():
        queue = asyncio.Queue()
        async def asker(a, b):
            l.append('start')
            queue.put_nowait(1)
            await queue.join()
            l.append('done')
            return Relation.cmp(a, b)
        async def other():
            await queue.get()
            l.append('in other')
            queue.task_done()
        return (await asyncio.gather(
            other(),
            avda(
                criteria = criteria,
                alts = alts,
                asker = asker,
                find_best = 1)))[1]

    assert (await f()).maxes(alts) == {(1, 0)}
    assert l == ['start', 'in other', 'done']

def asker_stub(a, b):
    raise ValueError

def test_one_criterion():
    '''When there's only one criterion, we should never need to ask
    questions, because the rule of dominance suffices to infer all
    preferences.'''

    for criterion_length in range(2, 10):
        prefs = artiruno.vda(
           criteria = [tuple(range(criterion_length))],
           asker = asker_stub)
        for i, j in choose2(range(criterion_length)):
            assert prefs.cmp((i,), (j,)) == LT

def test_dominant_maximum():
    '''Under `find_best = 1`, if there's an alternative that dominates
    all the others, we shouldn't need to ask any questions.'''

    criteria = tuple(tuple(range(5)) for _ in range(4))
    alts = [
        (2, 3, 1, 4),
        (2, 1, 3, 3),
        (2, 4, 1, 3),
        (3, 0, 2, 3),
        (4, 4, 3, 4)]
    prefs = artiruno.vda(criteria, alts, asker_stub, find_best = 1)
    assert prefs.maxes(among = alts) == {(4, 4, 3, 4)}

def test_big_item_space():
    n_criteria = 10
    n_levels = 20

    criteria = [list(range(n_levels))] * n_criteria
    alts = [[17] * n_criteria for _ in range(2)]
    alts[0][3] = 15
    alts[0][6] = 14
    alts[1][7] = 8
    alts[1][9] = 4

    prefs = vda(
        criteria = criteria,
        alts = alts,
        asker = Relation.cmp,
        find_best = True,
        max_dev = 3)
    assert prefs.cmp(tuple(alts[0]), tuple(alts[1])) == LT

def all_choice_seqs(
        criteria, alts = None, find_best = 1,
        max_dev = 2):

    choices = (LT, EQ, GT)
    if alts:
        alts = tuple(map(tuple, alts))
    queue = [(c,) for c in choices]
    result = []

    def asker(a, b):
        result[-1]['questions'].append((a, b))
        i = len(result[-1]['questions']) - 1
        if i == len(queue[0]):
            for c in choices[1:]:
                queue.append(queue[0] + (c,))
            queue[0] += (choices[0],)
        return queue[0][i]

    while queue:
        result.append(dict(questions = []))
        prefs = vda(
            criteria = criteria,
            alts = alts,
            asker = asker,
            find_best = find_best,
            max_dev = max_dev)
        result[-1]['choices'] = queue[0]
        result[-1]['prefs'] = prefs
        result[-1]['maxes'] = prefs.maxes(alts)
        queue.pop(0)

    return result

@pytest.mark.slow
def test_recode_criteria():
    '''Renaming criterion values shouldn't change the questions asked
    or the results.'''

    criteria = tuple(
        tuple('abc'[i] + str(j) for j in range(3)) for i in range(3))
    alts = (
        ('a2', 'b2', 'c0'),
        ('a2', 'b0', 'c2'),
        ('a1', 'b2', 'c1'),
        ('a0', 'b1', 'c2'))

    r1 = all_choice_seqs(criteria, alts, max_dev = 5)

    crev = tuple(c[::-1] for c in criteria)
    def revitem(item):
        return tuple(crev[c][criteria[c].index(v)] for c, v in enumerate(item))
    r2 = all_choice_seqs(
        criteria = crev,
        alts = tuple(map(revitem, alts)),
        max_dev = 5)

    assert len(r1) == len(r2)
    for d1, d2 in zip(r1, r2):
        assert d2['questions'] == [(revitem(a), revitem(b))
            for a, b in d1['questions']]
        assert d2['maxes'] == set(map(revitem, d1['maxes']))

@pytest.mark.slow
def test_irrelevant_criteria():
    for d in all_choice_seqs(
            criteria = ('abc', 'pqr', 'xyz'),
            find_best = None):
        def p(a, b):
            return d['prefs'].cmp(tuple(a), tuple(b))
        assert p('aqx', 'apy') == p('bqx', 'bpy') == p('cqx', 'cpy')

def test_lex_small():
    'A scenario with lexicographic preferences.'
    criteria = [(0, 1)] * 3
    alts = [(1, 0, 0), (0, 1, 1)]
    prefs = vda(
        criteria = criteria, alts = alts,
        asker = Relation.cmp,
        max_dev = 3)
    assert prefs.cmp(*alts) == Relation.cmp(*alts)

def test_lex_big():

    criteria = [(0, 1, 2)] * 3
    alts = [
      # I think I got these items from a paper, but now I've forgotten
      # which. D'oh.
        (0, 1, 1),
        (1, 1, 0),
        (2, 0, 1),
        (0, 2, 0),
        (1, 0, 2),
        (2, 1, 1),
        (2, 1, 0),
        (2, 2, 0),
        (1, 0, 1)]

    def p(find_best, max_dev, a): return vda(
        criteria = criteria,
        alts = alts,
        asker = a,
        find_best = find_best,
        max_dev = max_dev)

    prefs = p(1, 3, lambda a, b: Relation.cmp(a[::-1], b[::-1]))
    assert prefs.maxes() == {(2, 2, 2)}
    assert prefs.maxes(among = alts) == {(1, 0, 2)}

    prefs = p(1, 3, lambda a, b: Relation.cmp(a, b))
    assert prefs.maxes() == {(2, 2, 2)}
    assert prefs.maxes(among = alts) == {(2, 2, 0)}

    prefs = p(3, 4, lambda a, b: Relation.cmp(a, b))
    assert prefs.extreme(3, among = alts) == {
        (2, 2, 0), (2, 1, 1), (2, 1, 0)}

def random_scenarios(f):
    '''For each of several trials, randomly generate alternatives, and
    call `f` to randomly generate an asker. Check that the preferences
    Artiruno concludes are consistent with the asker.'''

    trials = 20
    criteria_fast = [(2,2), (3,4), (5,5), (2,2,2), (3,2,2), (3,3,3)]
    criteria_slow = [(2,) * 5, (5,) * 5]
    criteria_all = criteria_fast + criteria_slow

    @pytest.mark.parametrize('criteria', criteria_all)
    def out(criteria, run_slow_tests):
        if criteria in criteria_slow and not run_slow_tests:
            pytest.skip()

        questions_asked = 0
        expect_questions_asked = (
            inspect.signature(f).parameters['n_questions'].default)[
                criteria_all.index(criteria)]

        criteria = tuple(tuple(range(n)) for n in criteria)
        item_space = tuple(itertools.product(*criteria))

        for trial in range(trials):
            R = random.Random(repr((criteria, trial)))
            alts = R.sample(item_space,
                min(len(item_space), R.randint(2, 8)))
            find_best = R.choice([True, None])
            if find_best is True:
                find_best = R.choice(
                    [i for i in [1, 2, 3] if i < len(alts)])
            asker = f(criteria, R)
            def counting_asker(a, b):
                nonlocal questions_asked
                questions_asked += 1
                return asker(a, b)
            prefs = vda(
                criteria, alts, counting_asker, find_best,
                max_dev = 2*len(criteria))
            if find_best:
                assert prefs.extreme(find_best, among = alts) == {
                    a
                    for a in alts
                    for cmps in [Counter(asker(a, b) for b in alts)]
                    if cmps[IC] == 0 and cmps[LT] < find_best}
            else:
                for a, b in choose2(alts):
                    assert prefs.cmp(a, b) == asker(a, b)

        assert questions_asked == expect_questions_asked
           # If questions_asked < expect_questions_asked, that's an
           # improvement, so update expect_questions_asked.
           # If questions_asked > expect_questions_asked, that's a
           # regression, so fix it.

    return out

@random_scenarios
def test_lex_generalized(criteria, R,
        n_questions = (15, 39, 54, 34, 47, 80, 105, 642)):
    '''Artiruno should be able to reproduce lexicographic preferences,
    in which the criteria have a defined order of importance.'''
    criterion_order = R.sample(range(len(criteria)), len(criteria))
    return lambda a, b: Relation.cmp(*(
        tuple(v[i] for i in criterion_order)
        for v in (a, b)))

@random_scenarios
def test_value_function(criteria, R,
        n_questions = (15, 50, 53, 35, 43, 94, 72, 905)):
    '''Artiruno should be able to reproduce preferences defined by an
    additive value function, in which each criterion increment adds a
    certain positive amount of utility.'''
    values = tuple(
       tuple(0 if i == 0 else R.randint(1, 4)
           for i in range(len(c)))
       for c in criteria)
    return lambda a, b: Relation.cmp(*(
       sum(sum(values[ci][: cl + 1])
           for ci, cl in enumerate(item))
       for item in (a, b)))



'''This module provides a simple terminal-based interactive interface
to VDA. The session is initialized from the command line, and then
the user's choices are read from standard input.'''

import sys, json
from artiruno.preorder import LT, EQ, GT
from artiruno.vda import vda, Abort
from artiruno._version import __version__

def interact(criterion_names, alts, alt_names, **kwargs):

    def asker(a, b):
        print('\nWhich do you prefer?')

        options = dict(a = GT, b = LT, e = EQ, q = Abort)
        for k in 'a', 'b':
            item = dict(a = a, b = b)[k]
            print(f'\n({k})' + (
                ' <<< ' + alt_names[alts.index(item)]
                    if alt_names and item in alts else ''))
            for i, (name, value) in enumerate(zip(criterion_names, item)):
                print(f'- {name}: {value}' +
                    ('   <<< different' if a[i] != b[i] else ''))
        print('\n(e) The two options are equally preferable')
        print('(q) Abort')

        print('')
        while True:
            try:
                inp = input('> ').strip().lower()
            except (EOFError, KeyboardInterrupt):
                print('Aborted.')
                raise Abort()
            try:
                v = options[inp]
            except KeyError:
                print('Invalid input.')
                continue
            if v is Abort:
                print('Aborted.')
                raise Abort()
            return v

    return vda(
        asker = asker,
        alts = alts,
        allowed_pairs_callback = apc,
        **kwargs)

def apc(allowed_pairs):
    if len(allowed_pairs) > 1:
        print('Allowed pairs now:', allowed_pairs[0])

def setup_interactive(scenario):
    alt_names = None
    if alts := scenario.get('alts'):
        if isinstance(alts, dict):
            alt_names, alts = zip(*alts.items())
        alts = tuple(tuple(a[c] for c in scenario['criteria'])
            for a in alts)
    namer = ((lambda a: alt_names[alts.index(a)])
        if alt_names
        else str)
    interact_args = dict(
        criterion_names = list(scenario['criteria']),
        criteria = list(scenario['criteria'].values()),
        alt_names = alt_names,
        alts = alts,
        find_best = scenario.get('find_best'),
        max_dev = 2 * len(scenario['criteria']))
    return interact_args, alts, namer

def results_text(scenario, prefs, alts, namer):
    return (
        'Best: ' + ', '.join(namer(a)
            for a in prefs.extreme(scenario['find_best'], alts))
        if scenario.get('find_best') else
        'Preferences: ' + prefs.get_subset(alts).summary(namer))

def main():
    import argparse
    args = argparse.ArgumentParser(
        prog = 'artiruno',
        description = __doc__)
    args.add_argument('--version', action = 'version',
        version = 'Artiruno ' + __version__)
    args.add_argument('FILEPATH',
        help = 'path to a JSON file describing the scenario')
    args = args.parse_args()

    with open(args.FILEPATH) as o:
        scenario = json.load(o)

    interact_args, alts, namer = setup_interactive(scenario)
    prefs = interact(**interact_args)
    print(results_text(scenario, prefs, alts, namer))

    try:
        import graphviz
    except ModuleNotFoundError:
        print('The Python module `graphviz` is not installed. '
            'No graph for you.')
        return
    from tempfile import mktemp
    (prefs.get_subset(alts)
        .graph(namer = namer)
        .render(filename = mktemp(), format = 'png', view = True))



import setuptools, ast

with open('README.rst', 'r') as o:
    long_description = o.read()

with open('artiruno/_version.py', 'r') as o:
    version = ast.literal_eval(ast.parse(o.read()).body[0].value)

setuptools.setup(
    name = 'artiruno',
    version = version,
    author = 'Kodi B. Arfer',
    description = 'Verbal decision analysis',
    long_description = long_description,
    long_description_content_type = 'text/x-rst',
    url = 'https://arfer.net/projects/artiruno',
    project_urls = {
        "Documentation": 'https://arfer.net/projects/artiruno/doc',
        "Source Code": 'https://github.com/Kodiologist/Artiruno'},
    packages = setuptools.find_packages(),
    classifiers = [
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics'],
    python_requires = '>=3.8')



# This module is only run by Pyodide.

import asyncio, json
import js, pyodide
from artiruno.vda import avda
from artiruno.interactive import setup_interactive, results_text
from artiruno.preorder import LT, GT, EQ

# ------------------------------------------------------------
# * Global state
# ------------------------------------------------------------

vda_running = False
n_questions = 0

# ------------------------------------------------------------
# * Helpers
# ------------------------------------------------------------

class Quit(Exception): pass

T = js.document.createTextNode

signals = dict(
    choice = asyncio.Event(),
    quit_done = asyncio.Event())

def signal(event_type, value = True):
    signals[event_type].value = value
    signals[event_type].set()

async def get_signal(event_type):
    await signals[event_type].wait()
    signals[event_type].clear()
    value = signals[event_type].value
    signals[event_type].value = None
    return value

def E(id_str):
    return js.document.getElementById(id_str)

class H:
    'Create an HTML element, set attributes, and add children.'
    def __getattr__(self, element_name):
        def f(*children, **attrs):
            x = js.document.createElement(element_name)
            for c in children:
                x.appendChild(c)
            for k, v in attrs.items():
                x.setAttribute(k.lower(), v)
            return x
        setattr(H, element_name, f)
        return f
H = H()

# ------------------------------------------------------------
# * Main functions
# ------------------------------------------------------------

def initialize_web_interface():
    with open('examples/jobs.json', 'r') as o:
        E('problem-definition').value = o.read().strip()
    E('start-button-parent').appendChild(
        H.BUTTON(T('Start decision-making'), id = 'start-button'))
    E('start-button').addEventListener('click',
        pyodide.create_proxy(lambda _:
            asyncio.ensure_future(restart_decision_making())))

async def restart_decision_making():
    try:
       await _restart_decision_making()
    except Exception as e:
        E('dm').textContent = repr(e)
        return

async def _restart_decision_making():
    global vda_running

    # Terminate any current VDA and clear the log.
    if vda_running:
        # Kill the current VDA job.
        signal('choice', 'quit')
        # Wait till it's dead before continuing.
        await get_signal('quit_done')
    E('dm').textContent = ''

    # Reword "start" to "restart".
    E('start-button').textContent = 'Restart decision making'

    # Parse the problem definition.
    scenario = json.loads(E('problem-definition').value)
    interact_args, alts, namer = setup_interactive(scenario)

    # Begin VDA.
    try:
        vda_running = True
        prefs = await interact(**interact_args)
        E('dm').appendChild(H.P(T(
            results_text(scenario, prefs, alts, namer))))
    except Quit:
        signal('quit_done')
    finally:
        vda_running = False

async def interact(criterion_names, alts, alt_names, **kwargs):
    global n_questions
    n_questions = 0

    async def asker(a, b):
        # Present the choices as a list with buttons.

        global n_questions
        n_questions += 1

        buttons = {}
        def button(the_id, text):
            buttons[the_id] = pyodide.create_proxy(
                lambda _: signal('choice', the_id))
            return H.BUTTON(T(text), id = the_id)

        def display_item(item):
            # Display the item as a list with one criterion and value
            # per list item. Highlight the criteria value that differ
            # between the two options.
            return H.UL(*(
                H.LI(
                    T(name + ': '),
                    (T(value) if a[i] == b[i] else H.STRONG(T(value))))
                for i, (name, value)
                in enumerate(zip(criterion_names, item))))

        E('dm').appendChild(H.DIV(
            H.P(T('Q{}: Which would you prefer?'.format(n_questions))),
            H.UL(
                H.LI(button('option_a', 'Option A'),
                    display_item(a)),
                H.LI(button('option_b', 'Option B'),
                    display_item(b)),
                H.LI(button('equal', 'Equal'),
                    T('The two options are equally preferable'))),
            Class = 'query'))

        for button, callback in buttons.items():
            E(button).addEventListener('click', callback)

        # Wait for the user to click a button.
        choice = await get_signal('choice')
        if choice == 'quit':
            raise Quit()

        # Replace the buttons with indicators of the user's decision.
        for button, callback in buttons.items():
            E(button).replaceWith(H.SPAN(
               T('your choice' if button == choice else 'not chosen'),
               Class = 'chosen' if button == choice else 'not-chosen'))
            callback.destroy()

        # Return the choice.
        return dict(option_a = GT, option_b = LT, equal = EQ)[choice]

    return await avda(
        asker = asker,
        alts = alts,
        **kwargs)
